/**
  *  \file client/widgets/navwidget.cc
  *  \brief Class WShipNavigationChart
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include <algorithm>
#include "client/widgets/navwidget.h"
#include "game/host.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/obj.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "game/univ.h"
#include "util/nls.h"
#include "game/objlist.h"
#include "client/dlg-visual.h"
#include "client/screens.h"
#include "ui/widgets/abstractbutton.h"
#include "game/bbox.h"


/** Constructor. */
WShipNavigationChart::WShipNavigationChart(GUniverse& univ)
    : old_pos_head(0),
      old_pos_count(0),
      chunnel_mode(false),
      chunnel_button(0),
      chunnel_ship(0),
      circle_hyp(false),
      circle_radius(0),
      univ(univ),
      alt_circles(0, 0)
{ }

/** Draw warp circles at given position.
    \param can Canvas to operate on
    \param pt Game universe location */
void
WShipNavigationChart::drawCircles(GfxCanvas& can, GPoint pt)
{
    GChartViewport& vp = getViewport();
    GfxPoint p = vp.scaleXY(pt);
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_DARK);
    if (circle_hyp) {
        // xxxship
        // if (ship.isHyperdriving()) {
        // HYP circle
        drawCircle(ctx, p, vp.scale(340));
        drawCircle(ctx, p, vp.scale(360));
    } else if (circle_radius) {
        // ship.getWarp() >= 5) {
        // Warp circles
        drawCircle(ctx, p, vp.scale(circle_radius));
        drawCircle(ctx, p, vp.scale(2*circle_radius));
        if (vp.divi >= 2*vp.mult) {
            drawCircle(ctx, p, vp.scale(3*circle_radius));
        }
    } else {
        // no circles
    }
}

void
WShipNavigationChart::setCircleRadius(bool hyp, int r)
{
    circle_hyp = hyp;
    circle_radius = r;
}

void
WShipNavigationChart::setChunnelShip(const GShip& ship)
{
    chunnel_ship = &ship;
}

/** Set initial zoom.
    Defines a zoom level so that the initial scanner and the hyp circle,
    if any, fits into the view. */
void
WShipNavigationChart::setInitialZoom()
{
    GPoint scanner = getScannerTarget() - getScannerOrigin();
    GfxPoint size = getExtent().getSize();

    int min_dx = std::abs(scanner.x);
    int min_dy = std::abs(scanner.y);
    if (circle_hyp) {
        min_dx = std::max(360, min_dx);
        min_dy = std::max(360, min_dy);
    }

    int min_zoom = 1;
    if (min_dx > 0 && size.x > 0) {
        min_zoom = std::max(min_zoom,
                            std::min(int(GChartBBox::MAX_ZOOM),
                                     (2*min_dx + size.x-1) / size.x));
    }
    if (min_dy > 0 && size.y > 0) {
        min_zoom = std::max(min_zoom,
                            std::min(int(GChartBBox::MAX_ZOOM),
                                     (2*min_dy + size.y-1) / size.y));
    }
    getViewport().mult = 1;
    getViewport().divi = min_zoom;
}

void
WShipNavigationChart::drawPre(GfxCanvas& can)
{
    if (chunnel_mode) {
        /* Chunnel mode */
        GChartViewport& vp = getViewport();
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);

        GPoint center = getCenterPosition();

        /* Forbidden area */
        int dist = host.isPHost() ? config.MinimumChunnelDistance() : 100;
        if (dist > 0) {
            GfxPoint pt = vp.scaleXY(getScannerOrigin().getSimpleNearestAlias(center));
            int r = vp.scale(dist);
            ctx.setFillPattern(LTSLASH_FILL);
            ctx.setColor(COLOR_FIRE+4);
            drawFilledCircle(ctx, pt, r);
            drawCircle(ctx, pt, r);
        }

        /* Possible targets */
        ctx.setColor(COLOR_BRIGHTCYAN);  /* PCC 1.x uses GREENSCALE+15 */
        int r = vp.scale(10);
        if (r < 5)
            r = 5;
        if (r > 20)
            r = 20;
        for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
            const GShip& mate = univ.getShip(i);
            if (isValidChunnelMate(*chunnel_ship, mate)) {
                GfxPoint pt = vp.scaleXY(mate.getPos().getSimpleNearestAlias(center));
                int rr;
                if (chunnel_ship->getOwner() == mate.getOwner()) {
                    ctx.setColor(COLOR_BRIGHTCYAN);
                    rr = r;
                } else {
                    ctx.setColor(COLOR_DARKCYAN);
                    rr = r+1;
                }
                ctx.setCursor(GfxPoint(pt.x - rr, pt.y));
                drawLineRel(ctx, rr, -rr);
                drawLineRel(ctx, rr, rr);
                drawLineRel(ctx, -rr, rr);
                drawLineRel(ctx, -rr, -rr);
            }
        }
    } else {
        /* Regular mode, warp circles */
        drawCircles(can, getScannerOrigin());
        if (alt_circles.x > 0) {
            drawCircles(can, alt_circles);
        }
    }
}

bool
WShipNavigationChart::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_TAB:
            // Push position
            old_pos[old_pos_head] = getCenterPosition();
            old_pos_head = (old_pos_head + 1) % MAX_OLD_POS;
            if (old_pos_count < MAX_OLD_POS)
                ++old_pos_count;

            // Go there
            setCenterPosition(getScannerTarget());
            return true;

         case SDLK_BACKSPACE:
            if (old_pos_count > 0) {
                // Pop position
                old_pos_head = (old_pos_head + MAX_OLD_POS-1) % MAX_OLD_POS;
                --old_pos_count;
                setCenterPosition(old_pos[old_pos_head]);
            }
            return true;

         case SDLK_INSERT:
         case 'y':
            // Add circle
            if (!chunnel_mode) {
                alt_circles = getScannerTarget();
                drawWidget();
            }
            return true;

         case SDLK_DELETE:
            // Remove circle
            if (!chunnel_mode) {
                alt_circles = GPoint(0, 0);
                drawWidget();
            }
            return true;

         case 'l':
         case 'L':
            // List ships
         {
             GObjectReference ref = doListShips(getScannerTarget(), ls_IncludeForeignShips | ls_AllowForeignShips, 0, _("List Ships"), _("OK"), 0);
             if (ref.isValid()) {
                 sig_ship_select.raise(ref->getId());
             }
             return true;
         }

         case SDLK_F5:
         case ss_Ctrl + SDLK_F5:
            // Planet info
            if (int pid = univ.getPlanetAt(getScannerTarget(), true)) {
                doPlanetScan(pid);
            }
            return true;

         case ss_Shift + SDLK_F5:
            // Planet info here
            if (int pid = univ.getPlanetAt(getCenterPosition(), false)) {
                doPlanetScan(pid);
            }
            return true;
        }
    }
    return WScannerChartWidget::handleEvent(event, second_pass);
}

void
WShipNavigationChart::setChunnelButton(UIAbstractButton* btn)
{
    chunnel_button = btn;
    if (chunnel_button != 0) {
        chunnel_button->setState(UIAbstractButton::bf_Highlight, chunnel_mode);
    }
}

void
WShipNavigationChart::toggleChunnel()
{
    if (chunnel_ship != 0) {
        chunnel_mode = !chunnel_mode;
        setChunnelButton(chunnel_button);    // updates the button's status
        drawWidget();
    }
}

bool
WShipNavigationChart::isChunnelMode() const
{
    return chunnel_mode;
}

/** Perform lock on target. Called whenever the user presses
    Space/Return or a mouse button to lock onto an object. This
    overrides the behaviour to lock only on Fireclouds in chunnel
    mode. */
void
WShipNavigationChart::doItemLock(GPoint clicked, TLockMode mode, bool marked_only, bool optimize_warp)
{
    if (chunnel_mode) {
        const GChartViewport& vp = getViewport();
        GLockData locker(clicked,
                         GPoint(vp.unscaleX(extent.x),
                                vp.unscaleY(extent.y + extent.h - 1)),
                         GPoint(vp.unscaleX(extent.x + extent.w - 1),
                                vp.unscaleY(extent.y)),
                         marked_only);
        for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
            const GShip& mate = univ.getShip(i);
            if (isValidChunnelMate(*chunnel_ship, mate)) {
                locker.addObject(mate);
            }
        }
        userMoveScannerTo(locker.getPoint());
    } else {
        WScannerChartWidget::doItemLock(clicked, mode, marked_only, optimize_warp);
    }
}
