/**
  *  \file game/attachmentunpacker.cc
  *  \brief Attachment Unpacker
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <stdexcept>
#include "game/attachmentunpacker.h"
#include "cpluslib/buffer.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "cpluslib/utf8.h"
#include "game/rst.h"
#include "game/utildata.h"
#include "game/utilread.h"
#include "util/console.h"
#include "util/misc.h"
#include "util/nls.h"

// namespace gt = game::v3::structures;

namespace {
    struct Map {
        const char* tpl;
        GAttachmentUnpacker::Kind kind;
    };

    /* File name mapping:

       Everything not listed here is NormalFile (e.g. xtrfcode.txt, hullspec.dat).

       A CriticalFile is not unpacked by default.
       We need to guess whether a file is used legitimately.
       An attempt to overwrite a file that comes with a result normally is definitely fishy.
       Updating spec files can be fishy, but is also legitimate.
       If in doubt, we assume that host is somehow trusted.
       (They could as well send the file updates within a .zip.)


       Placeholders:

       "*" matches anything up to the next "." ("*foo" will not work),
       "#" matches any sequence of numbers.

       We do not match numeric placeholders to the current player number;
       although player3.rst never contains a ship4.dat, we still reject that. */
    const Map KIND_MAP[] = {
        { "*.bat",            GAttachmentUnpacker::CriticalFile },         // also blacklisted by VPA
        { "*.com",            GAttachmentUnpacker::CriticalFile },         // also blacklisted by VPA
        { "*.exe",            GAttachmentUnpacker::CriticalFile },         // also blacklisted by VPA
        { "*.ovr",            GAttachmentUnpacker::CriticalFile },         // also blacklisted by VPA
        { "*.q",              GAttachmentUnpacker::CriticalFile },
        { "*.qc",             GAttachmentUnpacker::CriticalFile },
        { "*.src",            GAttachmentUnpacker::ConfigurationFile },
        { "*.sys",            GAttachmentUnpacker::CriticalFile },         // also blacklisted by VPA
        { "bdata#.dat",       GAttachmentUnpacker::CriticalFile },
        { "bdata#.dis",       GAttachmentUnpacker::CriticalFile },
        { "cmd#.txt",         GAttachmentUnpacker::CriticalFile },
        { "config#.cc",       GAttachmentUnpacker::CriticalFile },
        { "contrl#.dat",      GAttachmentUnpacker::CriticalFile },
        { "control.dat",      GAttachmentUnpacker::CriticalFile },
        { "cp#.cc",           GAttachmentUnpacker::CriticalFile },
        { "fizz.bin",         GAttachmentUnpacker::CriticalFile },
        { "gen#.dat",         GAttachmentUnpacker::CriticalFile },
        { "hconfig.hst",      GAttachmentUnpacker::ConfigurationFile },
        { "kore#.dat",        GAttachmentUnpacker::CriticalFile },
        { "mdata#.dat",       GAttachmentUnpacker::CriticalFile },
        { "mess#.dat",        GAttachmentUnpacker::CriticalFile },
        { "mess35#.dat",      GAttachmentUnpacker::CriticalFile },
        { "pcc2.ini",         GAttachmentUnpacker::CriticalFile },
        { "pdata#.dat",       GAttachmentUnpacker::CriticalFile },
        { "pdata#.dis",       GAttachmentUnpacker::CriticalFile },
        { "player#.rst",      GAttachmentUnpacker::CriticalFile },
        { "player#.trn",      GAttachmentUnpacker::CriticalFile },
        { "race.nm",          GAttachmentUnpacker::RaceNameFile },
        { "ship#.dat",        GAttachmentUnpacker::CriticalFile },
        { "ship#.dis",        GAttachmentUnpacker::CriticalFile },
        { "shiplist.txt",     GAttachmentUnpacker::ConfigurationFile },
        { "shipxy#.dat",      GAttachmentUnpacker::CriticalFile },
        { "skore#.dat",       GAttachmentUnpacker::CriticalFile },
        { "target#.dat",      GAttachmentUnpacker::CriticalFile },
        { "target#.ext",      GAttachmentUnpacker::CriticalFile },
        { "util#.dat",        GAttachmentUnpacker::CriticalFile },
        { "vcr#.dat",         GAttachmentUnpacker::CriticalFile },
    };

    bool matchName(const string_t& name, const char* tpl)
    {
        // util::StringParser p(name);
        string_t::size_type p = 0;
        while (char ch = *tpl++) {
            if (ch == '*') {
                string_t::size_type n = name.rfind('.');
                if (n == string_t::npos || n < p) {
                    p = name.size();
                } else {
                    p = n;
                }
            } else if (ch == '#') {
                bool ok = false;
                while (p < name.size() && name[p] >= '0' && name[p] <= '9') {
                    ++p;
                    ok = true;
                }
                if (!ok) {
                    return false;
                }
            } else {
                if (p >= name.size() || name[p] != ch) {
                    return false;
                }
                ++p;
            }
        }
        return p == name.size();
    }

    /* Get default kind for a file, given its name. */
    GAttachmentUnpacker::Kind getFileKind(const string_t& name)
    {
        for (size_t i = 0; i < countof(KIND_MAP); ++i) {
            if (matchName(name, KIND_MAP[i].tpl)) {
                return KIND_MAP[i].kind;
            }
        }
        return GAttachmentUnpacker::NormalFile;
    }

    /* Check for invalid characters in a file name.
       We reject anything that looks like it contains a directory separator,
       and characters that are not file-system safe (i.e. non-ASCII).
       For now, we also reject space which shouldn't appear in VGAP files. */
    bool hasInvalidCharacter(const string_t& name)
    {
        for (size_t i = 0, n = name.size(); i < n; ++i) {
            uint8_t ch = static_cast<uint8_t>(name[i]);
            if (ch == '\\' || ch == '/' || ch == ':' || ch >= 0x80 || ch <= ' ') {
                return true;
            }
        }
        return false;
    }

    string_t canonicalizeFileName(string_t name)
    {
        return strTrim(strLCase(name));
    }

    /* Check whether file has given content. */
    bool hasSameContent(Directory& dir, const string_t& name, const Buffer<char>& content)
    {
        try {
            Ptr<Stream> s = dir.openFile(name, Stream::C_READ);
            size_t content_pos = 0;
            while (1) {
                // Read a page
                char buffer[4096];
                size_t read = s->read(buffer, sizeof(buffer));

                // If input ends, files are equal if expected content also ends
                if (read == 0) {
                    return content_pos >= content.getSize();
                }

                // Compare data. If expected content ends before input, this will yield false due to mismatching length.
                if (read > content.getSize() - content_pos) {
                    return false;
                }
                if (std::memcmp(&content[content_pos], buffer, read) != 0) {
                    return false;
                }
                content_pos += read;
            }
        }
        catch (...) {
            return false;
        }
    }

    /* Check for valid racename file.
       It seems the race.nm section in a result file is sometimes blank; avoid damaging user's files. */
    bool isValidRaceNames(const char buffer[], size_t n)
    {
        for (size_t i = 0; i < n; ++i) {
            if (uint8_t(buffer[i]) > 32) {
                return true;
            }
        }
        return false;
    }

    string_t sanitizeString(string_t in)
    {
        string_t result;
        Utf8Reader rdr(in);
        while (rdr.hasMore()) {
            uint32_t ch = rdr.get();
            if (ch >= ' ' && ch < 0x7F) {
                result += char(ch);
            } else {
                result += '?';
            }
        }
        return result;
    }

    const char NULL_TIMESTAMP[18] = {'0','0','-','0','0','-','0','0','0','0','0','0',':','0','0',':','0','0'};
}

/*
 *  Representation of an Attachment
 */

struct GAttachmentUnpacker::Attachment {
    string_t name;
    Kind kind;
    bool enabled;
    Buffer<char> content;
};


/*
 *  util.dat Reader for Attachment Extraction
 */

class GAttachmentUnpacker::Reader : public GUtilReader {
 public:
    Reader(GAttachmentUnpacker& parent)
        : m_parent(parent), m_openAttachment(0)
        { }
    virtual bool process(int rec_id, size_t size, const char data[])
        {
            switch (rec_id) {
             case 27:
                // Raw pconfig.src
                if (Attachment* a = m_parent.createAttachment("pconfig.src")) {
                    a->content.clear();
                    a->content.ensureSize(size);
                    std::memcpy(a->content.ptr(), data, size);
                }
                break;

             case 34:
                // General file
                if (size >= TUtil34FTP::size) {
                    // Fetch header
                    TUtil34FTP header;
                    getStructure(data, header);

                    // Build attachment
                    if (Attachment* a = m_parent.createAttachment(getBasicString(header.name))) {
                        size_t attSize = size - header.size;
                        a->content.clear();
                        a->content.ensureSize(attSize);
                        std::memcpy(a->content.ptr(), data + header.size, attSize);
                    }
                }
                break;

             case 59:
                // Long file
                if (size >= TUtil59FTP::size) {
                    // Fetch header and file name
                    TUtil59FTP header;
                    getStructure(data, header);
                    size -= TUtil59FTP::size;
                    data += TUtil59FTP::size;

                    if (header.fileNameLength > size) {
                        console.write(LOG_DEBUG, _("Attachment record is truncated"));
                        break;
                    }

                    const string_t fileName = getBasicStringN(data, header.fileNameLength);
                    size -= header.fileNameLength;
                    data += header.fileNameLength;

                    // Sort it to its place
                    if ((header.flags & FTP_NOTFIRST) != 0) {
                        // It's not the first one, so there needs to be a matching open attachment
                        if (m_openAttachment == 0 || m_openAttachment->name != canonicalizeFileName(fileName)) {
                            console.write(LOG_DEBUG, format(_("Attachment \"%s\" is missing parts."), fileName));
                            break;
                        }
                    } else {
                        // It is the first one, so there shouldn't be an open one
                        closeAttachment();
                        if (Attachment* a = m_parent.createAttachment(fileName)) {
                            a->content.clear();
                            m_openAttachment = a;
                        }
                    }

                    // Append
                    if (m_openAttachment != 0) {
                        m_openAttachment->content.append(data, size);
                    }

                    // Finish attachment
                    if ((header.flags & FTP_NOTLAST) == 0) {
                        m_openAttachment = 0;
                    }
                }
                break;
            }
            return true;
        }
    virtual void failure(Stream& /*in*/)
        { }

    // Change: PCC2 has no handleEnd() callback, so this does not generate a warning in this case

 private:
    void closeAttachment()
        {
            if (m_openAttachment != 0) {
                console.write(LOG_DEBUG, format(_("Attachment \"%s\" is missing parts."), m_openAttachment->name));
                m_openAttachment = 0;
            }
        }

    GAttachmentUnpacker& m_parent;
    Attachment* m_openAttachment;
};



/*
 *  GAttachmentUnpacker
 */

GAttachmentUnpacker::GAttachmentUnpacker()
    : m_acceptableKinds(Kinds_t::fromInteger(-1) - CriticalFile),
      m_timestamp(NULL_TIMESTAMP),
      m_attachments()
{ }

GAttachmentUnpacker::~GAttachmentUnpacker()
{ }

void
GAttachmentUnpacker::setAcceptableKind(Kind k, bool enable)
{
    if (enable) {
        m_acceptableKinds += k;
    } else {
        m_acceptableKinds -= k;
    }
}

void
GAttachmentUnpacker::loadDirectory(Directory& dir, int playerNumber)
{
    Ptr<Stream> in = dir.openFileNT(format("player%d.rst", playerNumber), Stream::C_READ);
    if (in.ptr() != 0) {
        loadResultFile(*in, playerNumber);
    }

    in = dir.openFileNT(format("util%d.dat", playerNumber), Stream::C_READ);
    if (in.ptr() != 0) {
        loadUtilData(*in, playerNumber);
    }
}

void
GAttachmentUnpacker::loadResultFile(Stream& in, int playerNumber)
{
    GResultFile rst(in);

    // Check timestamp
    in.seek(rst.getSectionOffset(GResultFile::sec_Gen));
    TResultGen gen;
    getStructureT(in, gen);
    if (!checkTimestamp(gen.timestamp)) {
        return;
    }

    // Check for race.nm file in KORE section
    if (rst.haveSection(GResultFile::sec_Kore)) {
        int32_t offset = rst.getSectionOffset(GResultFile::sec_Kore);
        in.seek(offset
                + 500 * sizeof(TKoreMine)
                + 50 * sizeof(TKoreStorm)
                + 50 * sizeof(TKoreExplosion));

        char raceNameBuffer[682];
        in.readT(raceNameBuffer, sizeof(raceNameBuffer));

        if (isValidRaceNames(raceNameBuffer, sizeof(raceNameBuffer))) {
            if (Attachment* a = createAttachment("race.nm")) {
                a->content.clear();
                a->content.append(raceNameBuffer, sizeof(raceNameBuffer));
            }
        }
    }

    // Check for LEECH section (probably nobody uses this today)
    if (rst.haveSection(GResultFile::sec_Leech)) {
        in.seek(rst.getSectionOffset(GResultFile::sec_Leech));

        TInt32 packedSize;
        getStructureT(in, packedSize);

        int32_t size = packedSize.size;
        if (size >= 0) {
            if (Attachment* a = createAttachment(format("leech%d.dat", playerNumber))) {
                a->content.clear();
                a->content.ensureSize(size);
                in.readT(a->content.ptr(), size);
            }
        }
    }
}

void
GAttachmentUnpacker::loadUtilData(Stream& in, int /*playerNumber*/)
{
    GTimestamp ts;
    if (Reader::check(in, &ts)) {
        if (checkTimestamp(ts)) {
            Reader(*this).read(in);
        }
    }
}

void
GAttachmentUnpacker::dropUnchangedFiles(Directory& dir)
{
    size_t o = 0;
    for (size_t i = 0; i < m_attachments.size(); ++i) {
        Attachment* p = m_attachments[i];
        if (!hasSameContent(dir, p->name, p->content)) {
            m_attachments.swap_elems(i, o);
            ++o;
        } else {
            console.write(LOG_DEBUG, format(_("File \"%s\" is unchanged."), p->name));
        }
    }
    while (m_attachments.size() > o) {
        m_attachments.pop_back();
    }
}

void
GAttachmentUnpacker::dropUnselectedAttachments()
{
    size_t o = 0;
    for (size_t i = 0; i < m_attachments.size(); ++i) {
        Attachment* p = m_attachments[i];
        if (p->enabled) {
            m_attachments.swap_elems(i, o);
            ++o;
        }
    }
    while (m_attachments.size() > o) {
        m_attachments.pop_back();
    }
}

bool
GAttachmentUnpacker::saveFiles(Directory& dir)
{
    bool ok = true;
    for (size_t i = 0; i < m_attachments.size(); ++i) {
        Attachment* p = m_attachments[i];
        if (p->enabled) {
            try {
                // Erase, so we're not surprised if it's a symlink
                dir.eraseNT(p->name);

                // Write
                Ptr<Stream> s = dir.openFile(p->name, Stream::C_CREATE);
                if (p->content.getSize() != 0) {
                    s->writeT(&p->content[0], p->content.getSize());
                }

                console.write(LOG_INFO, format(_("Received attachment \"%s\"."), p->name));
            }
            catch (std::exception& e) {
                console.write(LOG_ERROR, format(_("Unable to write attachment \"%s\": %s"), p->name, e.what()));
                ok = false;
            }
        } else {
                console.write(LOG_DEBUG, format(_("Attachment \"%s\" ignored."), p->name));
        }
    }
    return ok;
}

size_t
GAttachmentUnpacker::getNumAttachments() const
{
    return m_attachments.size();
}

GAttachmentUnpacker::Attachment*
GAttachmentUnpacker::getAttachmentByIndex(size_t i) const
{
    if (i < m_attachments.size()) {
        return m_attachments[i];
    } else {
        return 0;
    }
}

GAttachmentUnpacker::Attachment*
GAttachmentUnpacker::getAttachmentByName(const string_t& name) const
{
    for (size_t i = 0, n = m_attachments.size(); i < n; ++i) {
        if (m_attachments[i]->name == name) {
            return m_attachments[i];
        }
    }
    return 0;
}

GAttachmentUnpacker::Kind
GAttachmentUnpacker::getAttachmentKind(Attachment* p) const
{
    return p != 0
        ? p->kind
        : NormalFile;
}

void
GAttachmentUnpacker::selectAttachment(Attachment* p, bool enable)
{
    if (p != 0) {
        p->enabled = enable;
    }
}

void
GAttachmentUnpacker::selectAttachmentsByKind(Kind k, bool enable)
{
    for (size_t i = 0, n = m_attachments.size(); i < n; ++i) {
        if (m_attachments[i]->kind == k) {
            m_attachments[i]->enabled = enable;
        }
    }
}

void
GAttachmentUnpacker::selectAllAttachments(bool enable)
{
    for (size_t i = 0, n = m_attachments.size(); i < n; ++i) {
        m_attachments[i]->enabled = enable;
    }
}

bool
GAttachmentUnpacker::isAttachmentSelected(Attachment* p) const
{
    return p != 0 && p->enabled;
}

string_t
GAttachmentUnpacker::getAttachmentName(Attachment* p) const
{
    return p != 0
        ? p->name
        : string_t();
}

size_t
GAttachmentUnpacker::getAttachmentSize(Attachment* p) const
{
    return p != 0
        ? p->content.getSize()
        : 0;
}

GAttachmentUnpacker::Kinds_t
GAttachmentUnpacker::getAllAttachmentKinds() const
{
    Kinds_t result;
    for (size_t i = 0, n = m_attachments.size(); i < n; ++i) {
        result += m_attachments[i]->kind;
    }
    return result;
}

const GTimestamp&
GAttachmentUnpacker::getTimestamp() const
{
    return m_timestamp;
}

string_t
GAttachmentUnpacker::toString(Kind k)
{
    string_t result;
    switch (k) {
     case NormalFile:        result = _("File");               break;
     case ConfigurationFile: result = _("Configuration File"); break;
     case RaceNameFile:      result = _("Race Names");         break;
     case CriticalFile:      result = _("Dangerous File");     break;
    }
    return result;
}

GAttachmentUnpacker::Attachment*
GAttachmentUnpacker::createAttachment(string_t name)
{
    // Canonicalize name
    name = canonicalizeFileName(name);

    // Refuse invalid names
    if (name.empty() || name[0] == '.' || hasInvalidCharacter(name)) {
        console.write(LOG_DEBUG, format(_("Attachment \"%s\" ignored because of invalid file name."), sanitizeString(name)));
        return 0;
    }

    // Look for existing attachment (or insert position)
    size_t i = 0;
    while (i < m_attachments.size() && strCollate(name, m_attachments[i]->name) > 0) {
        ++i;
    }

    // Existing attachment?
    if (i < m_attachments.size() && name == m_attachments[i]->name) {
        return m_attachments[i];
    }

    // Create new attachment
    Attachment* na = m_attachments.insert(m_attachments.begin() + i, new Attachment());
    na->name = name;
    na->kind = getFileKind(name);
    na->enabled = m_acceptableKinds.contains(na->kind);
    return na;
}

bool
GAttachmentUnpacker::checkTimestamp(const GTimestamp& ts)
{
    if (ts == NULL_TIMESTAMP || (m_timestamp != NULL_TIMESTAMP && ts.isEarlierThan(m_timestamp))) {
        return false;
    }

    if (ts != m_timestamp) {
        m_attachments.clear();
        m_timestamp = ts;
    }
    return true;
}
