/**
  *  \file int/compound.h
  *  \brief Interpreter: Compound Values
  *
  *  This defines compound values, i.e. values which are not basic
  *  types of our language.
  *
  *  IntCallableValue is the basic type for an item which can be
  *  called. Derived from it is IntIndexableValue, which can appear as
  *  a function call "value(args)" on the left or right side of an
  *  assignment, or in iteration "ForEach value". 
  *  IntSimpleIndexableValue implements the common case of such a
  *  value, which is a global function with no associated data. 
  *  Likewise, IntSimpleProcedureValue implements the common case of a
  *  global procedure.
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_COMPOUND_H
#define PCC_V2_INT_COMPOUND_H

#include "cpluslib/smartptr.h"
#include "int/value.h"
#include "int/data.h"

class IntContext;
class IntExecutionContext;
class IntArgBlock;

/** Callable value. This is the base for items callable in a process context.
    They can have the syntactic form of a procedure or of a function. */
class IntCallableValue : public IntValue {
 public:
    /** Call. This implements invocation of a subroutine, in the form "a := value(args)".
        Subroutines without return value have to return null instead.
        \param exc   Execution context
        \param args  Data segment containing parameters
        \param want_result True if a result is required (use exc.pushNewValue()) */
    virtual void call(IntExecutionContext& exc, IntDataSegment& args, bool want_result) = 0;

    /** Check syntactic form. */
    virtual bool isProcedureCall() = 0;

    /** Array reflection. Implementation of the IsArray() and Dim() builtins.
        \param which 0=get number of dimensions, 1..n=get that dimension */
    virtual int32_t getDimension(int32_t which) = 0;

    /** Get context for first item in iteration. This should be equivalent to
        <tt>dynamic_cast<IntContext*>(call(exc, args))</tt>, where
        args are the parameters needed to address the first object.
        Caller assumes lifetime management for the context.
        \return context. Return 0 if set is empty. Throws IntError if request is invalid. */
    virtual IntContext* makeFirstContext() = 0;
};

/** Indexable value. This value can be used in two forms, "value(args)" and
    "ForEach value Do ...". */
class IntIndexableValue : public IntCallableValue {
 public:
    /** Call. This implements "a := value(args)".
        \param dseg  Data segment containing parameters
        \param index Index of first parameter
        \param nargs Number of parameters
        \return obtained value. Throws IntError if request is invalid */
    virtual IntValue* get(IntArgBlock& args) = 0;

    /** Set indexed value. This implements "value(args) := a".
        \param dseg  Data segment containing parameters
        \param index Index of first parameter
        \param nargs Number of parameters
        \param value Value to assign.
        \return obtained value. Throws IntError if request is invalid */
    virtual void set(IntArgBlock& args, IntValue* value) = 0;

    bool isProcedureCall();
    void call(IntExecutionContext& exc, IntDataSegment& args, bool want_result);
};

/** Simple indexable value. Implements an IntIndexableValue supporting read access
    and iteration, based upon two function pointers. In other words, this implements
    a global function, with possible iteration access. */
class IntSimpleIndexableValue : public IntIndexableValue {
 public:
    typedef int32_t     (*dim_t)(int32_t);
    typedef IntValue*   (*get_t)(IntArgBlock&);
    typedef IntContext* (*make_t)();

    IntSimpleIndexableValue(dim_t dim, get_t get, make_t make);
    ~IntSimpleIndexableValue();

    // IntIndexableValue:
    IntValue*   get(IntArgBlock& args);
    void        set(IntArgBlock& args, IntValue* value);
    int32_t     getDimension(int32_t dim);
    IntContext* makeFirstContext();

    // IntValue:
    string_t    toString(bool readable);
    void        store(IntTagNode&, Stream&);
    IntValue*   clone();

 private:
    const dim_t  dim_function;
    const get_t  get_function;
    const make_t make_function;
};

/** Simple procedure value. Implements an IntCallableValue based upon a function
    pointer, handed to userland as a procedure. */
class IntSimpleProcedureValue : public IntCallableValue {
 public:
    typedef void (*call_t)(IntExecutionContext& exc, IntArgBlock& args);

    IntSimpleProcedureValue(call_t call);
    ~IntSimpleProcedureValue();

    // IntCallableValue:
    void      call(IntExecutionContext& exc, IntDataSegment& args, bool want_result);
    bool      isProcedureCall();
    int32_t   getDimension(int32_t);
    IntContext* makeFirstContext();

    // IntValue:
    string_t    toString(bool readable);
    void        store(IntTagNode&, Stream&);
    IntValue*   clone();

 private:
    call_t  call_function;
};

/** Closure. */
class IntClosure : public IntCallableValue {
 public:
    IntClosure();
    ~IntClosure();

    // IntClosure:
    void setNewFunction(IntCallableValue* function);
    void addNewArgument(IntValue* value);
    void addNewArgumentsFrom(IntDataSegment& seg, uint32_t nargs);

    // IntCallableValue:
    virtual void call(IntExecutionContext& exc, IntDataSegment& args, bool want_result);
    virtual bool isProcedureCall();
    virtual int32_t getDimension(int32_t which);
    virtual IntContext* makeFirstContext();

    // IntValue:
    virtual string_t toString(bool readable);
    virtual void store(IntTagNode&, Stream&);
    virtual IntValue* clone();

 private:
    IntClosure(const IntClosure& other);

    // Both attributes are smart pointers, so we can share them without cloning.
    // They are not modified.
    Ptr<IntCallableValue> function;
    Ptr<IntDataSegment> fixed_args;
};

#endif
