/**
  *  \file int/globalif.h
  *  \brief Interpreter Game Interface: Global Functions and Properties
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_GLOBALIF_H
#define PCC_V2_INT_GLOBALIF_H

#include "cpluslib/types.h"
#include "int/context.h"
#include "int/compound.h"

class IntValue;
class IntArgBlock;
class IntDataSegment;
class IntExecutionContext;
class GGen;

class IntGlobalContext : public IntSingleContext {
 public:
    IntGlobalContext();
    ~IntGlobalContext();

    // IntContext:
    int32_t   lookup(const IntNameQuery& name);
    void      set(int32_t index, IntValue* value);
    IntValue* get(int32_t index);
    IntContext* clone();
    GObject*  getObject();

    string_t  toString(bool readable);
    void      store(IntTagNode& sv, Stream& aux);
    void      enumProperties(IntPropertyAcceptor& acceptor);
};

/** Implementation of CC$NotifyConfirmed(). This must be implemented
    separately using the full interface (not the simplified one)
    because it needs an execution context. */
class IntCCNotifyConfirmed : public IntCallableValue {
 public:
    IntCCNotifyConfirmed();
    ~IntCCNotifyConfirmed();

    // IntCallableValue:
    void call(IntExecutionContext& exc, IntDataSegment& args, bool want_result);
    bool isProcedureCall();
    int32_t getDimension(int32_t);
    IntContext* makeFirstContext();

    // IntValue:
    string_t toString(bool readable);
    void store(IntTagNode& sv, Stream& aux);
    IntCCNotifyConfirmed* clone();
};


enum IntGlobalProperty {
    igpFileFormatLocal,
    igpFileFormatRemote,
    igpGameDirectory,
    igpMyInMsgs,
    igpMyOutMsgs,
    igpMyVCRs,
    igpRootDirectory,
    igpSelectionLayer,
    igpSystemLanguage,
    igpSystemProgram,
    igpSystemVersion,
    igpSystemVersionCode,
    igpSystemHost,
    igpSystemHostCode,
    igpSystemHostVersion,
    igpRandomSeed,
    igpRegSharewareFlag,
    igpRegSharewareText,
    igpRegStr1,
    igpRegStr2,
    igpTurnNumber,
    igpTurnDate,
    igpTurnIsNew,
    igpTurnTime
};

IntValue* getGlobalProperty(IntGlobalProperty igp);
void      setGlobalProperty(IntGlobalProperty igp, IntValue* value);

bool checkSavedPassword(int32_t player, const GGen& gen);
void clearSavedPasswords();

IntValue* IFDistanceGet(IntArgBlock& args);
IntValue* IFShipNameGet(IntArgBlock& args);
IntValue* IFPlanetNameGet(IntArgBlock& args);
IntValue* IFPlanetAtGet(IntArgBlock& args);
IntValue* IFCfgGet(IntArgBlock& args);
IntValue* IFPrefGet(IntArgBlock& args);
IntValue* IFIsSpecialFCodeGet(IntArgBlock& args);
IntValue* IFTruehull(IntArgBlock& args);
IntValue* IFRandomFCode(IntArgBlock& args);
IntValue* IFFormat(IntArgBlock& args);
IntValue* IFTranslate(IntArgBlock& args);
IntValue* IFRandom(IntArgBlock& args);

IntValue* IFCAdd(IntArgBlock& args);
IntValue* IFCCompare(IntArgBlock& args);
IntValue* IFCDiv(IntArgBlock& args);
IntValue* IFCExtract(IntArgBlock& args);
IntValue* IFCMul(IntArgBlock& args);
IntValue* IFCRemove(IntArgBlock& args);
IntValue* IFCSub(IntArgBlock& args);

void IFAddCommand(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFAddConfig(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFAddFCode(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFAuthPlayer(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFNewCircle(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFNewRectangle(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFNewRectangleRaw(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFNewLine(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFNewLineRaw(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFNewMarker(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFSaveGame(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFSelectionLoad(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFSelectionSave(IntExecutionContext& /*exc*/, IntArgBlock& args);

void IFCCSelectionExec(IntExecutionContext& /*exc*/, IntArgBlock& args);
void IFCCNotify(IntExecutionContext& exc, IntArgBlock& args);

#endif
