/**
  *  \file int/notify.h
  *  \brief Notification Messages
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_NOTIFY_H
#define PCC_V2_INT_NOTIFY_H

#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "cpluslib/ptrvector.h"
#include "cpluslib/sigslot.h"
#include "game/msg.h"

class IntExecutionContext;
class IntNotificationMessageStore;

/** Notification message. Notification messages can be unconfirmed or confirmed.
    They can be associated with a process. Each process can have up to one
    notification message and can query its status.

    IntNotificationMessage's lifetime is managed by IntNotificationMessageStore.
    All IntNotificationMessage objects are stored in an IntNotificationMessageStore,
    usually its global instance. */
class IntNotificationMessage {
 private:
    friend class IntNotificationMessageStore;
    IntNotificationMessage(string_t header, string_t body, IntNotificationMessageStore& store);

 public:
    ~IntNotificationMessage();
    string_t getText() const;
    const string_t& getHeader() const;
    const string_t& getBody() const;

    void setAssociatedProcess(IntExecutionContext* process);
    IntExecutionContext* getAssociatedProcess() const;

    bool isConfirmed() const;
    void setConfirmed(bool value);

    void remove();

 private:
    string_t header, body;
    bool confirmed;
    IntExecutionContext* associated_process;
    IntNotificationMessageStore& store;
};


/** Container for notification messages. */
class IntNotificationMessageStore : public GMessagebox {
 public:
    typedef uint32_t index_t;
    static const index_t nil = index_t(-1);

    IntNotificationMessageStore();
    ~IntNotificationMessageStore();

    index_t getNumMessages() const;
    index_t getIndexOfMessage(IntNotificationMessage* msg);

    IntNotificationMessage* getMessageByIndex(index_t index);
    IntNotificationMessage* addNewMessage(string_t header, string_t body);

    void removeMessageByIndex(index_t idx);
    void makeConfirmedProcessesRunnable();

    // GMessagebox:
    int getCount();
    string_t getText(int index);
    string_t getHeading(int index);

    Signal0 sig_changed;

    static IntNotificationMessageStore instance;

 private:
    ptr_vector<IntNotificationMessage> messages;
};



inline void
IntNotificationMessage::setAssociatedProcess(IntExecutionContext* process)
{
    associated_process = process;
}

inline IntExecutionContext*
IntNotificationMessage::getAssociatedProcess() const
{
    return associated_process;
}

inline bool
IntNotificationMessage::isConfirmed() const
{
    return confirmed;
}

#endif
