/**
  *  \file tools/c2plugin.cc
  *  \brief c2plugin - Plugin Manager
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <cstdlib>
#include "util/modinit.h"
#include "arch/platform.h"
#include "util/pluginmgr.h"
#include "util/plugin.h"
#include "cpluslib/format.h"
#include "util/console.h"
#include "util/nls.h"
#include "version.h"
#include "cpluslib/misc.h"
#include "util/except.h"
#include "util/progress.h"
#include "util/pluginstall.h"
#include "io/nullstream.h"

/******************************* Utilities *******************************/

const char* progname;

static void
die(string_t str)
{
    console.write(LOG_ERROR, format("%s: %s", progname, str));
    std::exit(1);
}

static void
createPath(string_t name)
{
    string_t parent = Directory::getDirName(name);
    string_t child  = Directory::getFileName(name);

    try {
        Ptr<Directory> d = Directory::openDir(parent);
        Ptr<DirectoryEntry> e = d->getEntryByName(child);
        if (e->getFileType() != DirectoryEntry::FT_DIRECTORY)
            e->createAsDirectory();
    }
    catch (FileProblemException& e) {
        console.write(LOG_ERROR, format("%s: %s", e.getFileName(), e.what()));
    }
}

static bool
checkRemovePlugin(PluginManager& mgr, Plugin& plug)
{
    std::vector<Plugin*> tmp;
    mgr.enumDependingPlugins(plug, tmp);
    if (!tmp.empty()) {
        console.write(LOG_ERROR, format(_("Plugin '%s' is required by the following plugins:"), plug.getId()));
        for (size_t i = 0, n = tmp.size(); i < n; ++i) {
            console.write(LOG_ERROR, format("  %s (%s)", tmp[i]->getId(), tmp[i]->getName()));
        }
        console.write(LOG_ERROR, _("It cannot be uninstalled."));
        return false;
    }
    return true;
}

static bool
checkPreconditions(PluginManager& mgr, Plugin& plug)
{
    // Check for conflicts
    std::vector<Plugin*> plugList;
    mgr.enumConflictingPlugins(plug, plugList);
    if (!plugList.empty()) {
        console.write(LOG_ERROR, format(_("Plugin '%s' conflicts with the following plugins:"), plug.getId()));
        for (size_t i = 0, n = plugList.size(); i < n; ++i) {
            console.write(LOG_ERROR, format("  %s (%s)", plugList[i]->getId(), plugList[i]->getName()));
        }
        console.write(LOG_ERROR, _("It cannot be installed."));
        return false;
    }

    // Check for preconditions
    Plugin::FeatureSet fset;
    mgr.enumFeatures(fset);
    if (!plug.isSatisfied(fset)) {
        console.write(LOG_ERROR, format(_("Plugin '%s' requires the following features:"), plug.getId()));
        Plugin::FeatureSet missing;
        plug.enumMissingFeatures(fset, missing);
        for (Plugin::FeatureSet::const_iterator it = missing.begin(), e = missing.end(); it != e; ++it) {
            string_t txt = it->first;
            if (!it->second.empty()) {
                txt += " ";
                txt += it->second;
            }
            console.write(LOG_ERROR, "  " + txt);
        }
        console.write(LOG_ERROR, _("It cannot be installed."));
        return false;
    }
    return true;
}

/****************************** Test Dummies *****************************/

namespace {
    class NullDirectory : public Directory {
     public:
        NullDirectory()
            { }
        ~NullDirectory()
            { }
        Ptr<DirectoryEntry> getEntryByName(string_t name)
            { return new Entry(this, name); }
        Ptr<Enumeration<Ptr<DirectoryEntry> > > getEntries()
            { return new Enum(); }
        Ptr<Directory> getParent()
            { return new NullDirectory(); }
        string_t getName()
            { return string_t(); }
     private:
        class Enum : public Enumeration<Ptr<DirectoryEntry> > {
         public:
            Ptr<DirectoryEntry> getNextElement()
                { return 0; }
            bool hasMoreElements()
                { return false; }
        };

        class Entry : public DirectoryEntry {
         public:
            Entry(Ptr<Directory> parent, string_t name)
                : DirectoryEntry(name),
                  parent(parent)
                { }
            Ptr<Directory> openDir()
                { return new NullDirectory(); }
            Ptr<Stream> openFile(int mode)
                {
                    if (mode & Stream::C_CREATE) {
                        return new NullStream();
                    } else {
                        throw FileProblemException(getName(), _("No such file"));
                    }
                }
            Ptr<Directory> openParent()
                { return parent; }
            void erase()
                { }
            void renameTo(string_t /*name*/)
                { }
            void createAsDirectory()
                { }
            void getInfo()
                { }
         private:
            Ptr<Directory> parent;
        };
    };
}


/******************************** Commands *******************************/

static void
doList(char** argv)
{
    // We do not take arguments
    enum { Default, Long, Short } f = Default;
    bool ordered = false;
    while (const char* p = *argv++) {
        if (*p == '-') {
            while (const char ch = *++p) {
                switch (ch) {
                 case 'l': f = Long; break;
                 case 'b': f = Short; break;
                 case 'o': ordered = true; break;
                 default:
                 {
                     char str[3] = {'-',ch,0};
                     die(format(_("Unknown option `%s'"), str));
                 }
                }
            }
        } else {
            die(_("too many arguments"));
        }
    }

    // Create plugin manager
    PluginManager& mgr = PluginManager::getInstance();
    mgr.findPlugins(Directory::makeFileName(Platform::getUserSettingsDirectory(1), "plugins"));

    // List them
    std::vector<Plugin*> them;
    mgr.enumPlugins(them, ordered);
    if (f != Short) {
        console << format(_("%d plugin%!1{s%} installed.\n"), them.size());
    }
    for (size_t i = 0, n = them.size(); i < n; ++i) {
        const Plugin& p = *them[i];
        if (f == Short) {
            console << p.getId() << "\n";
        } else {
            console << "--------\n"
                    << format(_("Plugin '%s': %s\n"), p.getId(), p.getName());
            if (!p.getDescription().empty()) {
                console << "\n" << p.getDescription() << "\n";
            }

            if (f == Long) {
                const Plugin::ItemList& items = p.getItems();
                bool did = false;
                for (size_t ii = 0, nn = items.size(); ii < nn; ++ii) {
                    if (items[ii].type != Plugin::Command) {
                        if (!did) {
                            console << "\n" << format(_("Files (in '%s'):"), p.getBaseDirectory()) << "\n";
                            did = true;
                        }
                        console << "  " << items[ii].name << "\n";
                    }
                }
            }
        }
    }
}

static void
doAdd(char** argv)
{
    // Create plugin directory
    string_t profileDirectory = Platform::getUserSettingsDirectory(1);
    string_t pluginDirName = Directory::makeFileName(profileDirectory, "plugins");
    createPath(profileDirectory);
    createPath(pluginDirName);

    // Create plugin manager
    PluginManager& mgr = PluginManager::getInstance();
    Ptr<Directory> dir = Directory::openDir(pluginDirName);
    mgr.findPlugins(*dir);

    ProgressMonitor pm;
    PluginInstaller installer(mgr, *dir, pm);

    // Iterate
    bool dry = false;
    bool did = false;
    bool err = false;
    bool force = false;
    while (char* p = *argv++) {
        if (std::strcmp(p, "-n") == 0) {
            dry = true;
        } else if (std::strcmp(p, "-f") == 0) {
            force = true;
        } else if (*p == '-') {
            die(format(_("Unknown option `%s'"), p));
        } else {
            did = true;
            try {
                Plugin* plug = installer.prepareInstall(p);
                if (!plug) {
                    console.write(LOG_ERROR, format(_("%s: Unknown file type"), p));
                    err = true;
                } else {
                    if (mgr.getPluginById(plug->getId())) {
                        console.write(LOG_INFO, format(_("Updating plugin '%s'..."), plug->getId()));
                    } else {
                        console.write(LOG_INFO, format(_("Installing plugin '%s'..."), plug->getId()));
                    }
                    if (force || checkPreconditions(mgr, *plug)) {
                        installer.doInstall(dry);
                    } else {
                        err = true;
                    }
                }
            }
            catch (FileProblemException& e) {
                console.write(LOG_ERROR, format("%s: %s", e.getFileName(), e.what()));
                err = true;
            }
            catch (std::exception& e) {
                console.write(LOG_ERROR, e.what());
                err = true;
            }
        }
    }
    if (!did) {
        die(format(_("Missing name plugin or file to install. '%s -h' for help."), progname));
    }
    if (err) {
        std::exit(1);
    }
}

static void
doRemove(char** argv)
{
    // Create plugin manager
    PluginManager& mgr = PluginManager::getInstance();
    Ptr<Directory> dir = Directory::openDir(Directory::makeFileName(Platform::getUserSettingsDirectory(1), "plugins"));
    mgr.findPlugins(*dir);

    ProgressMonitor pm;
    PluginInstaller installer(mgr, *dir, pm);

    // Iterate
    bool dry = false;
    bool did = false;
    bool err = false;
    bool force = false;
    while (char* p = *argv++) {
        if (std::strcmp(p, "-n") == 0) {
            dry = true;
        } else if (std::strcmp(p, "-f") == 0) {
            force = true;
        } else if (*p == '-') {
            die(format(_("Unknown option `%s'"), p));
        } else {
            did = true;
            if (Plugin* pPlug = mgr.getPluginById(strUCase(p))) {
                if (force || checkRemovePlugin(mgr, *pPlug)) {
                    console.write(LOG_INFO, format(_("Removing plugin '%s'..."), pPlug->getId()));
                    installer.doRemove(pPlug, dry);
                } else {
                    err = true;
                }
            } else {
                console.write(LOG_ERROR, format(_("Plugin '%s' is not known."), p));
                err = true;
            }
        }
    }
    if (!did) {
        die(format(_("Missing name of plugin to uninstall. '%s -h' for help."), progname));
    }
    if (err) {
        std::exit(1);
    }
}

static void
doTest(char** argv)
{
    // Dummy path
    Ptr<Directory> dir = new NullDirectory();

    // Iterate
    bool did = false;
    bool err = false;
    bool verbose = false;
    while (char* p = *argv++) {
        if (std::strcmp(p, "-v") == 0) {
            verbose = true;
        } else if (*p == '-') {
            die(format(_("Unknown option `%s'"), p));
        } else {
            // Create plugin manager and installer separately for each item
            PluginManager mgr;
            ProgressMonitor pm;
            PluginInstaller installer(mgr, *dir, pm);
            did = true;
            
            try {
                Plugin* plug = installer.prepareInstall(p);
                if (!plug) {
                    console.write(LOG_ERROR, format(_("%s: Unknown file type"), p));
                    err = true;
                } else {
                    installer.doInstall(false);
                    if (verbose) {
                        console.write(LOG_INFO, format(_("%s: Plugin '%s' (%s) tested successfully."), p, plug->getName(), plug->getId()));
                    }
                }
            }
            catch (FileProblemException& e) {
                if (p == e.getFileName()) {
                    console.write(LOG_ERROR, format("%s: %s", p, e.what()));
                } else {
                    console.write(LOG_ERROR, format("%s: %s: %s", p, e.getFileName(), e.what()));
                }
                err = true;
            }
            catch (std::exception& e) {
                console.write(LOG_ERROR, format("%s: %s", p, e.what()));
                err = true;
            }
        }
    }
    if (!did) {
        die(format(_("Missing name plugin or file to test. '%s -h' for help."), progname));
    }
    if (err) {
        std::exit(1);
    }
}

static void
doHelp(char** /*argv*/)
{
    console << format(_("PCC2 Plugin Manager v%s - (c) 2015-2021 Stefan Reuther\n"), PCC2_VERSION)
            << format(_("\n"
                        "Usage:\n"
                        "  %s -h|help\n"
                        "                  This help message\n"
                        "  %$0s list|ls [-l|-b] [-o]\n"
                        "                  List installed plugins\n"
                        "  %$0s add|install [-n] [-f] FILE.c2p...\n"
                        "                  Install given plugins\n"
                        "  %$0s remove|rm|uninstall [-n] [-f] ID...\n"
                        "                  Remove given plugins\n"
                        "  %$0s test [-v] FILE.c2p...\n"
                        "                  Test given plugins\n"
                        "\n"
                        "Options:\n"
                        " -l               List more details\n"
                        " -b               List fewer details\n"
                        " -o               List in load order (default: alphabetical)\n"
                        " -n               Dry run (don't modify anything, just check)\n"
                        " -f               Ignore dependencies/requirements\n"
                        " -v               Verbose\n"
                        "\n"
                        "Report bugs to <Streu@gmx.de>\n"),
                      progname);
    console.flush();
    std::exit(0);
}

/***************************** Initialisation ****************************/

namespace {
    struct Command {
        const char* name;
        void (*func)(char**);
    };

    static struct Command commands[] = {
        { "ls",        doList },
        { "list",      doList },
        { "add",       doAdd },
        { "install",   doAdd },
        { "rm",        doRemove },
        { "remove",    doRemove },
        { "uninstall", doRemove },
        { "test",      doTest },
        { "-h",        doHelp },
        { "--help",    doHelp },
        { "-help",     doHelp },
        { "help",      doHelp },
    };

    const Command* findCommand(const char* name)
    {
        for (size_t i = 0; i < countof(commands); ++i) {
            if (std::strcmp(commands[i].name, name) == 0) {
                return &commands[i];
            }
        }
        return 0;
    }

    const ModuleInitializer* c2plugin_modules[] = {
        mod_base_console,
        0
    };
}

int main(int /*argc*/, char** argv)
{
    // Initialize
    performInitialisation(OnStartup, c2plugin_modules);

    argv = Platform::convertArgumentsToUtf8(argv);
    progname = *argv++;

    // Check command
    const char* cmd = *argv;
    if (cmd == 0) {
        die(_("No command specified. Use 'c2plugin -h' for help."));
    } else if (const Command* p = findCommand(cmd)) {
        try {
            ++argv;
            p->func(argv);
        }
        catch (FileProblemException& e) {
            die(e.getFileName() + ": " + e.what());
        }
        catch (std::exception& e) {
            die(e.what());
        }
    } else {
        die(_("Invalid command specified. Use 'c2plugin -h' for help."));
    }
}
