/**
  *  \file ui/layout.cc
  *  \brief Layout Management
  *
  *  Layout managers arrange widgets in a group, without needing to
  *  manually compute coordinates. Layouts automatically adapt to
  *  changing font sizes and so forth, provided the widgets report
  *  correct values via getLayoutInfo().
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include <algorithm>
#include <vector>
#include "ui/layout.h"
#include "cpluslib/format.h"
#include "cpluslib/assert.h"

UIHBoxLayout UIHBoxLayout::instance0(0), UIHBoxLayout::instance5(5);
UIVBoxLayout UIVBoxLayout::instance0(0), UIVBoxLayout::instance5(5);

namespace {
    /** Information for HBox/VBox layouts. */
    struct LayoutInfo {
        std::vector<int> pref_sizes, min_sizes, flags;
        int used_space;
        int used_outer;

        LayoutInfo();
        std::vector<int>& doLayout(int space, int outer, int have_size, int flex_flag);
    };
};

LayoutInfo::LayoutInfo()
{ }

/** Compute generic layout.
    \pre pref_sizes, min_sizes, flags filled in
    \param space      inter-widget space we want to use
    \param outer      outer space we want to leave
    \param have_size  available space
    \param flex_flag  the flag signifying flexibility in our direction
                      (li_FlexibleH or li_FlexibleV)
    \returns vector of widget sizes
    \post used_outer = actual outer space to leave
    \post used_space = actual inter-widget space to leave */
std::vector<int>&
LayoutInfo::doLayout(int space, int outer, int have_size, int flex_flag)
{
    typedef std::vector<int>::size_type index;
    ASSERT(pref_sizes.size() == min_sizes.size());
    ASSERT(pref_sizes.size() == flags.size());

    /* First, figure out widget counts and sizes */
    int num_layout = 0, num_flex = 0;
    int total_pref = 0, total_min = 0;
    for (index i = 0; i < flags.size(); ++i) {
        if (flags[i] & UIBaseWidget::li_NoLayout) {
            ;
        } else {
            ++num_layout;
            total_pref += pref_sizes[i];
            total_min  += min_sizes[i];
            if (flags[i] & flex_flag)
                ++num_flex;
        }
    }

    /* Pathological case */
    if (!num_layout)
        return pref_sizes;

    /* Space that would be needed to layout this */
    used_space = space;
    if (num_layout > 1) {
        // too little space, reduce inter-widget gap
        while (used_space > 0 && (have_size < (num_layout-1) * used_space + total_min))
            --used_space;
        have_size -= (num_layout-1) * used_space;
    }

    used_outer = outer;
    while (used_outer > 0 && (have_size < 2*used_outer + total_min))
        // too little space, reduce outer space
        --used_outer;
    have_size -= 2*used_outer;

    if (have_size < total_pref) {
        /* Total size too small. Scale down relatively using minimum sizes. */
        /* FIXME: we can probably do better for have_size \in [min_total, pref_total] */
        int total = have_size;
        for (index i = 0; i < min_sizes.size(); ++i) {
            if (flags[i] & UIBaseWidget::li_NoLayout) {
                min_sizes[i] = 0;
            } else {
                int new_size = total_min ? min_sizes[i] * total / total_min : 0;
                total     -= new_size;
                total_min -= min_sizes[i];
                min_sizes[i] = new_size;
            }
        }
        ASSERT(total_min == 0);
        ASSERT(total == 0);
        return min_sizes;
    } else if (have_size > total_pref) {
        /* Total size larger than preferred size. Scale up using
           preferred sizes. */
        int delta = have_size - total_pref;  // total amount to enlarge
        if (num_flex) {
            /* we have flexible components; enlarge only those */
            for (index i = 0; i < pref_sizes.size(); ++i) {
                if (!(flags[i] & UIBaseWidget::li_NoLayout) && (flags[i] & flex_flag)) {
                    int d = delta / num_flex--;
                    pref_sizes[i] += d;
                    delta -= d;
                }
            }
            ASSERT(num_flex == 0);
            ASSERT(delta == 0);
        } else {
            /* no flexible: enlarge all components */
            for (index i = 0; i < pref_sizes.size(); ++i) {
                if (!(flags[i] & UIBaseWidget::li_NoLayout)) {
                    int d = delta / num_layout--;
                    pref_sizes[i] += d;
                    delta -= d;
                }
            }
            ASSERT(delta == 0);
            ASSERT(num_layout == 0);
        }
        return pref_sizes;
    } else
        return pref_sizes;
}

/****************************** UIHBoxLayout *****************************/

/** \class UIHBoxLayout
    \brief Horizontal Box Layout

    Widgets will be arranged horizontally, all the same height, from
    left to right. This layout will completely cover the container
    with widgets (subject to space/outer settings, of course). */

/** Create a Horizontal Box layout.
    \param space space to leave between widgets, in pixels.
    \param outer space to leave at left/right side. */
UIHBoxLayout::UIHBoxLayout(int space, int outer)
    : space(space), outer(outer)
{ }

void
UIHBoxLayout::getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info)
{
    int fland = UIBaseWidget::li_FlexibleH | UIBaseWidget::li_FlexibleV | UIBaseWidget::li_NoLayout;
    int flor  = 0;
    
    info.min_size = info.pref_size = GfxPoint(2*outer, 0);

    int n = 0;
    for (UIBaseWidget* w = container.getFirstChild(); w != 0; w = w->getNextSibling()) {
        UIBaseWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        fland &= winfo.flags;
        flor  |= winfo.flags;
        if (winfo.flags & UIBaseWidget::li_NoLayout)
            continue;
        ++n;
        info.min_size.x  += winfo.min_size.x;
        info.min_size.y   = std::max(info.min_size.y, winfo.min_size.y);
        info.pref_size.x += winfo.pref_size.x;
        info.pref_size.y  = std::max(info.pref_size.y, winfo.pref_size.y);
    }
    if (n != 0) {
        info.min_size.x  += (n-1) * space;
        info.pref_size.x += (n-1) * space;
    }

    /* A HBox is...
       ...FlexibleH if it has at least one component which is FlexibleH
          (if we have to enlarge the group, we can enlarge this component
          to compensate)
       ...FlexibleV if all components are FlexibleV (if one of them is not,
          we try not to squeeze it around)
       ...NoLayout if all components are NoLayout (only in this case we can
          ignore this group for layout) */
    info.flags = fland | (flor & UIBaseWidget::li_FlexibleH);
}

void
UIHBoxLayout::doLayout(UILayoutableWidget& container, GfxRect /*old_size*/, GfxRect new_size)
{
    LayoutInfo lay;
    for (UIBaseWidget* w = container.getFirstChild(); w; w = w->getNextSibling()) {
        UIBaseWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        lay.pref_sizes.push_back(winfo.pref_size.x);
        lay.min_sizes.push_back(winfo.min_size.x);
        lay.flags.push_back(winfo.flags);
    }
    if (new_size.w == 0)
        return;

    std::vector<int>& v = lay.doLayout(space, outer, new_size.w, UIBaseWidget::li_FlexibleH);
    doLayoutWithWidths(container, new_size, v, lay.used_space, lay.used_outer);
}

void
UIHBoxLayout::doLayoutWithWidths(UILayoutableWidget& container, GfxRect size,
                                 std::vector<int>& sizes, int space, int outer)
{
    std::vector<int>::size_type i = 0;
    size.x += outer;
    for (UIBaseWidget* w = container.getFirstChild(); w; w = w->getNextSibling(), ++i) {
        UIBaseWidget::LayoutInfo li;
        w->getLayoutInfo(li);
        if (li.flags & UIBaseWidget::li_NoLayout)
            continue;
        w->setExtent(GfxRect(size.x, size.y, sizes[i], size.h), true);
        size.x += sizes[i];
        size.x += space;
    }
}

/****************************** UIVBoxLayout *****************************/

/** \class UIVBoxLayout
    \brief Vertical Box Layout

    Widgets will be arranged vertically, all the same width, below
    each other from top to bottom. This layout will completely cover
    the container with widgets (subject to space/outer settings, of
    course). */

/** Create a Vertical Box layout.
    \param space space to leave between widgets, in pixels.
    \param outer space to leave at top/bottom */
UIVBoxLayout::UIVBoxLayout(int space, int outer)
    : space(space), outer(outer)
{ }

void
UIVBoxLayout::doLayout(UILayoutableWidget& container, GfxRect /*old_size*/, GfxRect new_size)
{
    LayoutInfo lay;
    for (UIBaseWidget* w = container.getFirstChild(); w; w = w->getNextSibling()) {
        UIBaseWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        lay.pref_sizes.push_back(winfo.pref_size.y);
        lay.min_sizes.push_back(winfo.min_size.y);
        lay.flags.push_back(winfo.flags);
    }
    if (new_size.h == 0)
        return;

    std::vector<int>& v = lay.doLayout(space, outer, new_size.h, UIBaseWidget::li_FlexibleV);
    doLayoutWithHeights(container, new_size, v, lay.used_space, lay.used_outer);
}

void
UIVBoxLayout::getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info)
{
    int fland = UIBaseWidget::li_FlexibleH | UIBaseWidget::li_FlexibleV | UIBaseWidget::li_NoLayout;
    int flor  = 0;

    info.min_size = info.pref_size = GfxPoint(0, 2*outer);

    int n = 0;
    for (UIBaseWidget* w = container.getFirstChild(); w != 0; w = w->getNextSibling()) {
        UIBaseWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        fland &= winfo.flags;
        flor  |= winfo.flags;
        if (winfo.flags & UIBaseWidget::li_NoLayout)
            continue;
        ++n;
        info.min_size.y  += winfo.min_size.y;
        info.min_size.x   = std::max(info.min_size.x, winfo.min_size.x);
        info.pref_size.y += winfo.pref_size.y;
        info.pref_size.x  = std::max(info.pref_size.x, winfo.pref_size.x);
    }
    if (n) {
        info.min_size.y  += (n-1) * space;
        info.pref_size.y += (n-1) * space;
    }

    info.flags = fland | (flor & UIBaseWidget::li_FlexibleV);
}

void
UIVBoxLayout::doLayoutWithHeights(UILayoutableWidget& container, GfxRect size,
                                  std::vector<int>& sizes, int space, int outer)
{
    std::vector<int>::size_type i = 0;
    size.y += outer;
    for (UIBaseWidget* w = container.getFirstChild(); w; w = w->getNextSibling(), ++i) {
        UIBaseWidget::LayoutInfo li;
        w->getLayoutInfo(li);
        if (li.flags & UIBaseWidget::li_NoLayout)
            continue;
        w->setExtent(GfxRect(size.x, size.y, size.w, sizes[i]), true);
        size.y += sizes[i];
        size.y += space;
    }
}

/****************************** UIGridLayout *****************************/

/*! \class UIGridLayout
  \brief Grid Layout (2D)

    This layout manager arranges widgets in a grid. The container will
    be divided with a grid, and each cell contains a widget. The cell
    widths and heights are computed automatically. */

/** Construct new grid.
    \param x      number of columns in the grid (number of rows
                  follows implicitly)
    \param space  space (pixels) between widgets, for both X and Y
                  direction
    \param outer  space (pixels) left left at the outside (all four
                  borders)

    Note that the layout parameters are "soft" restrictions and are
    lowered when space gets tight. */
UIGridLayout::UIGridLayout(int x, int space, int outer)
    : x(x), space(space), outer(outer),
      forced_cell_height(-1), forced_cell_width(-1)
{ }

static void
computeGridLayoutInfo(UIGridLayout& layout,
                      UILayoutableWidget& container,
                      LayoutInfo& hinfo,
                      LayoutInfo& vinfo,
                      const int x)
{
    int row = 0, col = 0;

    for (UIBaseWidget* w = container.getFirstChild(); w != 0; w = w->getNextSibling()) {
        UIBaseWidget::LayoutInfo info;
        w->getLayoutInfo(info);
        if (info.flags & UIBaseWidget::li_NoLayout)
            continue;

        // override with required values
        if (layout.getForcedCellWidth() >= 0) {
            info.pref_size.x = info.min_size.x = layout.getForcedCellWidth();
            info.flags &= ~UIBaseWidget::li_FlexibleH;
        }
        if (layout.getForcedCellHeight() >= 0) {
            info.pref_size.y = info.min_size.y = layout.getForcedCellHeight();
            info.flags &= ~UIBaseWidget::li_FlexibleV;
        }

        if (row == 0) {
            // first row, populate hinfo
            hinfo.pref_sizes.push_back(info.pref_size.x);
            hinfo.min_sizes.push_back(info.min_size.x);
            hinfo.flags.push_back(info.flags);
        } else {
            // not first row, update hinfo accordingly
            if (hinfo.pref_sizes[col] < info.pref_size.x)
                hinfo.pref_sizes[col] = info.pref_size.x;
            if (hinfo.min_sizes[col] < info.min_size.x)
                hinfo.min_sizes[col] = info.min_size.x;
            hinfo.flags[col] |= info.flags;
        }

        if (col == 0) {
            // first column, populate vinfo
            vinfo.pref_sizes.push_back(info.pref_size.y);
            vinfo.min_sizes.push_back(info.pref_size.y);
            vinfo.flags.push_back(info.flags);
        } else {
            if (vinfo.pref_sizes.back() < info.pref_size.y)
                vinfo.pref_sizes.back() = info.pref_size.y;
            if (vinfo.min_sizes.back() < info.min_size.y)
                vinfo.min_sizes.back() = info.min_size.y;
            vinfo.flags.back() |= info.flags;
        }

        if (++col >= x) {
            col = 0;
            ++row;
        }
    }
}

void
UIGridLayout::doLayout(UILayoutableWidget& container, GfxRect /*old_size*/, GfxRect new_size)
{
    LayoutInfo h, v;
    computeGridLayoutInfo(*this, container, h, v, x);

    if (!new_size.h || !new_size.w || h.min_sizes.empty())
        return;

    std::vector<int>& hsizes = h.doLayout(space, outer, new_size.w, UIBaseWidget::li_FlexibleH);
    std::vector<int>& vsizes = v.doLayout(space, outer, new_size.h, UIBaseWidget::li_FlexibleV);

    std::size_t row = 0, col = 0;
    int x = h.used_outer, y = v.used_outer;

    for (UIBaseWidget* p = container.getFirstChild(); p != 0; p = p->getNextSibling()) {
        UIBaseWidget::LayoutInfo li;
        p->getLayoutInfo(li);
        if (li.flags & UIBaseWidget::li_NoLayout)
            continue;

        ASSERT(col < hsizes.size() && row < vsizes.size());

        p->setExtent(GfxRect(new_size.x + x, new_size.y + y, hsizes[col], vsizes[row]), true);

        ++col;
        if (col >= size_t(this->x)) {
            x = h.used_outer;
            y += vsizes[row] + v.used_space;
            col = 0;
            ++row;
        } else {
            x += hsizes[col-1] + h.used_space;
        }
    }
}

void
UIGridLayout::getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info)
{
    LayoutInfo h, v;
    computeGridLayoutInfo(*this, container, h, v, x);

    if (h.min_sizes.empty()) {
        info.min_size = info.pref_size = GfxPoint(2 * outer, 2 * outer);
        info.flags = 0;
        return;
    }

    info.min_size = info.pref_size = GfxPoint(2 * outer + (h.min_sizes.size()-1) * space,
                                              2 * outer + (v.min_sizes.size()-1) * space);
    info.flags = 0;

    for (std::vector<int>::size_type i = 0; i < h.min_sizes.size(); ++i) {
        info.min_size.x += h.min_sizes[i];
        info.pref_size.x += h.pref_sizes[i];
        info.flags |= h.flags[i] & UIBaseWidget::li_FlexibleH;
    }
    for (std::vector<int>::size_type i = 0; i < v.min_sizes.size(); ++i) {
        info.min_size.y += v.min_sizes[i];
        info.pref_size.y += v.pref_sizes[i];
        info.flags |= v.flags[i] & UIBaseWidget::li_FlexibleV;
    }
}

/** Set forced cell size. You can force cells to have a particular
    size, with no respect to the contained widgets' wishes. By default,
    no requirements are given, thus the layout manager respects all
    widgets' layout requests.
    \param forced_cell_width required width for cells. -1 to accept any size.
    \param forced_cell_height required height for cells. -1 to accept any size. */
void
UIGridLayout::setForcedCellSize(int forced_cell_width, int forced_cell_height)
{
    this->forced_cell_width  = forced_cell_width;
    this->forced_cell_height = forced_cell_height;
}
